"""`rainbowio` module.

Provides the `colorwheel()` function."""

from __future__ import annotations

def colorwheel(n: float) -> int:
    """C implementation of the common colorwheel() function found in many examples.
    Returns the colorwheel RGB value as an integer value for n (usable in neopixel and dotstar).
    """
    ...
